from __future__ import annotations

import re
from itertools import count
from io import StringIO
import numpy as np


class Parameters(dict):
    parameter_counter = count()
    namespace: dict = {}

    def __init__(self):
        super().__init__()
        self._mapping: dict = {}

    def __str__(self) -> str:
        return 'Parameters:\n' + '\n'.join([f'{k}: {v}' for k, v in self.items()])

    def __getitem__(self, item) -> Parameter:
        if item in self._mapping:
            return super().__getitem__(self._mapping[item])
        else:
            return super().__getitem__(item)

    def __setitem__(self, key, value):
        super().__setitem__(key, value)

    def add(self,
            name: str,
            value: str | float | int = None,
            *,
            var: bool = True,
            lb: float = -np.inf, ub: float = np.inf) -> Parameter:

        par = Parameter(name=name, value=value, var=var, lb=lb, ub=ub)

        key = f'p{next(Parameters.parameter_counter)}'

        self.add_parameter(key, par)

        return par

    def add_parameter(self, key: str, parameter: Parameter):
        self._mapping[parameter.name] = key
        self[key] = parameter

        self._mapping[parameter.name] = key

        self[key] = parameter
        parameter.eval_allowed = False
        self.namespace[key] = parameter.value
        parameter.namespace = self.namespace
        parameter.eval_allowed = True

        # look for variables in expression and replace with valid names
        for p in self.values():
            if p._expr is not None:
                expression = p._expr
                for n, k in self._mapping.items():
                    expression = re.sub(re.escape(n), k, expression)

                p._expr = expression

    def replace_parameter(self, key_out: str, key_in: str, parameter: Parameter):
        for k, v in self._mapping.items():
            if v == key_out:
                self._mapping[k] = key_in
                break

        self.add_parameter(key_in, parameter)
        del self.namespace[key_out]

        for p in self.values():
            try:
                p.value
            except NameError:
                expression = p._expr_disp
                for n, k in self._mapping.items():
                    expression = re.sub(re.escape(n), k, expression)
                p._expr = expression

    def get_state(self):
        return {k: v.get_state() for k, v in self.items()}


class Parameter:
    """
    Container for one parameter
    """

    def __init__(self, name: str, value: float | str, var: bool = True, lb: float = -np.inf, ub: float = np.inf):
        self._value: float | None = None
        self.var: bool = bool(var) if var is not None else True
        self.error:  None | float = None if self.var is False else 0.0
        self.name: str = name
        self.function: str = ""

        self.lb: None | float = lb if lb is not None else -np.inf
        self.ub: float | None = ub if ub is not None else np.inf
        self.namespace: dict = {}
        self.eval_allowed: bool = True
        self._expr: None | str = None
        self._expr_disp: None | str = None

        if isinstance(value, str):
            self._expr_disp = value
            self._expr = value
            self.var = False
        else:
            if self.lb <= value <= self.ub:
                self._value = value
            else:
                raise ValueError('Value of parameter is outside bounds')

            self.init_val = value

            with np.errstate(divide='ignore'):
                # throws RuntimeWarning for zeros
                self.scale = 10**(np.floor(np.log10(np.abs(self.value))))

            if self.scale == 0:
                self.scale = 1.

    def __str__(self) -> str:
        start = StringIO()
        if self.name:
            if self.function:
                start.write(f"{self.name} ({self.function}):  ")
            else:
                start.write(self.name)
                start.write(":  ")

        if self.var:
            start.write(f"{self.value:.4g} +/- {self.error:.4g}, init={self.init_val}")
        else:
            start.write(f"{self.value:}")
            if self._expr is None:
                start.write(" (fixed)")
            else:
                start.write(f" (calc: {self._expr_disp})")

        return start.getvalue()

    def __add__(self, other: Parameter | float | int) -> float:
        if isinstance(other, (float, int)):
            return self.value + other
        elif isinstance(other, Parameter):
            return self.value + other.value

    def __radd__(self, other: Parameter | float) -> float:
        return self.__add__(other)

    @property
    def scaled_value(self) -> float:
        return self.value / self.scale

    @scaled_value.setter
    def scaled_value(self, value: float) -> None:
        self._value = value * self.scale

    @property
    def value(self) -> float:
        if self._expr is not None and self.eval_allowed:
            return eval(self._expr, {}, self.namespace)
        elif self._value is not None:
            return self._value

    @property
    def scaled_error(self) -> None | float:
        if self.error is None:
            return self.error
        else:
            return self.error / self.scale

    @scaled_error.setter
    def scaled_error(self, value) -> None:
        self.error = value * self.scale

    def get_state(self) -> dict:
        return {slot: getattr(self, slot) for slot in self.__slots__}

    @staticmethod
    def set_state(state: dict) -> Parameter:
        par = Parameter(state.pop('value'))
        for k, v in state.items():
            setattr(par, k, v)

        return par

    @property
    def full_name(self) -> str:
        name = self.name
        if self.function:
            name += f" ({self.function})"

        return name

    def copy(self) -> Parameter:
        if self._expr:
            val = self._expr_disp
        else:
            val = self._value
        para_copy = Parameter(name=self.name, value=val, var=self.var, lb=self.lb, ub=self.ub)
        para_copy._expr = self._expr
        para_copy.namespace = self.namespace

        return para_copy

